/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

#define _XOPEN_SOURCE	500

#include <stdio.h>
#include <stdlib.h>
#ifdef DEBUG_PEPC_PEPD
    #include <assert.h>
#endif /* DEBUG_PEPC_PEPD */
#include <string.h>

/* For usleep() */
#include <unistd.h>

#include "pep-c-interact.h"
#include "pep-c-obligation-handlers.h"
#include "pep-c-obligation-handlers_helpers.h"

#include "pep-c-log.h"

#include <lcmaps/lcmaps_log.h>


static xacml_request_t *  pepc_request   = NULL;
static xacml_response_t * pepc_response  = NULL;
static const char *prev_endpoint         = NULL;


/*
pep_subject_t * subject         = NULL;
pep_action_t * action           = NULL;
pep_resource_t * resource       = NULL;
pep_environment_t * environment = NULL;
*/



#if PEP_API == 1
int pepc_initialize(pep_obligationhandler_t **oh, size_t *oh_count) {
#else
PEP *pepc_initialize(pep_obligationhandler_t **oh, size_t *oh_count) {
#endif

    pep_obligationhandler_t   *oh_tmp = NULL;
    size_t                     i, j, oh_max;
    pep_error_t                pep_rc = PEP_OK;
   
    /* Make sure prev_endpoint is (re)initialised */
    prev_endpoint=NULL;

#if PEP_API == 1
    /* Make sure to initialize */
    *oh=NULL; *oh_count=0;

    /* Initialize the PEP-C library */
    lcmaps_log(LOG_DEBUG, "initialize PEP...\n");
    pep_rc= pep_initialize();
    if (pep_rc != PEP_OK) {
        lcmaps_log (LOG_ERR, "Error: pep_initialize() failed: %d: %s\n", pep_rc,pep_strerror(pep_rc));
        pep_rc= pep_destroy();
        if (pep_rc != PEP_OK) {
            lcmaps_log (LOG_ERR, "Error: pep_destroy() failed: %d: %s\n", pep_rc,pep_strerror(pep_rc));
            return pep_rc;
        }
        return -1;
    }
#else	/* PEP_API == 1 */
    PEP *pep_handle;

    /* Make sure to initialize */
    *oh=NULL; *oh_count=0;

    /* Initialize the PEP-C library */
    lcmaps_log(LOG_DEBUG, "initialize PEP...\n");
    pep_handle = pep_initialize();
    if (pep_handle == NULL) {
        lcmaps_log (LOG_ERR, "Error: pep_initialize() failed\n");
        pep_destroy(pep_handle);
        return NULL;
    }
#endif	/* PEP_API == 1 */

    /* Disable the PIPs since we don't need them */
#if PEP_API == 1
    pep_rc= pep_setoption(PEP_OPTION_ENABLE_PIPS,0);
    if (pep_rc != PEP_OK) {
        lcmaps_log (LOG_ERR,
		"Error: pep_setoption(PEP_OPTION_ENABLE_PIPS,0) failed: %s\n",
		pep_strerror(pep_rc));
        pep_destroy();
	return NULL;
    }
#else	/* PEP_API == 1 */
    pep_rc= pep_setoption(pep_handle, PEP_OPTION_ENABLE_PIPS,0);
    if (pep_rc != PEP_OK) {
        lcmaps_log (LOG_ERR,
		"Error: pep_setoption(pep_handle, PEP_OPTION_ENABLE_PIPS,0) failed: %s\n",
		pep_strerror(pep_rc));
        pep_destroy(pep_handle);
	return NULL;
    }
#endif	/* PEP_API == 1 */

    /* Register the obligation handlers presented by the registerObligationHandlers() function */
    lcmaps_log(LOG_DEBUG, "register Obligation Handlers: collect.\n");
    if ( registerObligationHandlers (&oh_tmp, &oh_max) != 0)
#if PEP_API == 1
	return -1;
#else
	return NULL;
#endif

    /* Loop the returned list, only if it makes sense */
    for (i = 0; (i < oh_max) && 
                ( oh_tmp[i].id) && 
                (strlen ( oh_tmp[i].id) > 0); i++)
    {
        lcmaps_log(LOG_DEBUG,
		"register Obligation Handlers: add obligation handler[%u] \"%s\" to pep-c.\n",
		(unsigned)i, oh_tmp[i].id);
#if PEP_API == 1
        pep_rc = pep_addobligationhandler( &(oh_tmp[i]) );
        if (pep_rc != PEP_OK) 
        {
            lcmaps_log (LOG_ERR,
		    "Error: pep_addobligationhandler() failed: %d: %s\n",
		    (int)pep_rc, pep_strerror(pep_rc));
            pep_destroy();
	    for (j=0; j<oh_max; j++) free( oh_tmp[j].id );
	    free(oh_tmp);
            return -1;
        }
#else
        pep_rc = pep_addobligationhandler(pep_handle, &(oh_tmp[i]) );
        if (pep_rc != PEP_OK) 
        {
            lcmaps_log (LOG_ERR,
		    "Error: pep_addobligationhandle() failed: %d: %s\n",
		    (int)pep_rc, pep_strerror(pep_rc));
            pep_destroy(pep_handle);
	    for (j=0; j<oh_max; j++) free( oh_tmp[j].id);
	    free(oh_tmp);
            return NULL;
        }
#endif
    }

    /* Explicitly enable the obligation handler triggering */
#if PEP_API == 1
    pep_rc= pep_setoption(PEP_OPTION_ENABLE_OBLIGATIONHANDLERS,1);
    if (pep_rc != PEP_OK) {
        lcmaps_log (LOG_ERR,
		"Error: pep_setoption(PEP_OPTION_ENABLE_OBLIGATIONHANDLERS,1) failed: %s\n",
		pep_strerror(pep_rc));
        pep_destroy();
	for (j=0; j<oh_max; j++) free(oh_tmp[j].id);
	free(oh_tmp);
        return -1;
    }
    *oh=oh_tmp; *oh_count=oh_max;
    return 0;
#else
    pep_rc= pep_setoption(pep_handle, PEP_OPTION_ENABLE_OBLIGATIONHANDLERS,1);
    if (pep_rc != PEP_OK) {
        lcmaps_log (LOG_ERR,
		"Error: pep_setoption(pep_handle, PEP_OPTION_ENABLE_OBLIGATIONHANDLERS,1) failed: %s\n",
		pep_strerror(pep_rc));
        pep_destroy(pep_handle);
	for (j=0; j<oh_max; j++) free(oh_tmp[j].id);
	free(oh_tmp);
        return NULL;
    }
    *oh=oh_tmp; *oh_count=oh_max;
    return pep_handle;
#endif
}

int pepc_construct_request (int use_profile,
                            const char * subject_proxy,
                            const char * actiontype,
                            const char * actionid,
                            const char * resourceid,
                            const char * resourcetype,
                            const char * dnshostname,
                            const char * environ_proxy)
{
    const char * logstr = "pepc_construct_request";

    lcmaps_log(LOG_DEBUG, "%s: create PEP request...\n", logstr);
    pepc_request = xacml_request_create();
    if (!pepc_request)
    {
        lcmaps_log (LOG_ERR, "%s: failed to create pepc_request object\n", logstr);
        return 1;
    }

    if (pepc_construct_request_subject (&pepc_request, 
                                        subject_proxy,
					use_profile))
    {
	lcmaps_log (LOG_ERR, "%s: failed to construct request subject object\n", logstr);
        return 1;
    }
    if (pepc_construct_request_action (&pepc_request, 
                                        actiontype,
                                        actionid))
    {
	lcmaps_log (LOG_ERR, "%s: failed to construct request action object\n", logstr);
        return 1;
    }
    if (pepc_construct_request_resource (&pepc_request, 
                                          resourceid, 
                                          resourcetype, 
                                          dnshostname,
					  use_profile))
    {
	lcmaps_log (LOG_ERR, "%s: failed to construct request resource object\n", logstr);
        return 1;
    }
    if (pepc_construct_request_environment(&pepc_request,
                                            environ_proxy,
					    use_profile))
    {
	lcmaps_log (LOG_ERR, "%s: failed to construct request environment object\n", logstr);
        return 1;
    }

    return 0;
}

int pepc_construct_request_subject    (xacml_request_t ** request, 
                                       const char * subject_proxy,
				       int use_profile)
{
    const char * logstr = "pepc_construct_request_subject";
    xacml_subject_t * subject = NULL;
    xacml_attribute_t * subj_attr = NULL;

    if ((!request) || (!(*request))) {
        lcmaps_log (LOG_ERR, "No request object presented, failure\n");
        return 1;
    }

    lcmaps_log(LOG_DEBUG, "add PEP subject(proxy)...\n");
    subject = xacml_subject_create();

    if (!subject) {
        lcmaps_log (LOG_ERR,
		"%s: failed to create xacml_subject_t object\n", logstr);
        return 1;
    }

    /* Add SWITCH Argus profile subject attribute */
    if (use_profile & PROFILE_SWITCH_GRIDWN)	{
	subj_attr=xacml_attribute_create(XACML_GRIDWN_SUBJECT_CERT_CHAIN);
	if (!subj_attr)	{
	    lcmaps_log (LOG_ERR,
		    "%s: failed to create xacml_attribute_t object\n", logstr);
	    return 1;
	}
	lcmaps_log(LOG_DEBUG, "  adding subject attribute %s\n",
		XACML_GRIDWN_SUBJECT_CERT_CHAIN);
	xacml_attribute_addvalue(subj_attr, subject_proxy);
	xacml_subject_addattribute(subject,subj_attr);
    }

    /* Add AuthZ Interop profile subject attribute */
    if (use_profile & PROFILE_AUTHZ_INTEROP)	{
	subj_attr=xacml_attribute_create(XACML_ATTR_PROFILE_NS_URL_CERT_CHAIN);
	if (!subj_attr)	{
	    lcmaps_log (LOG_ERR,
		    "%s: failed to create xacml_attribute_t object\n", logstr);
	    return 1;
	}
	lcmaps_log(LOG_DEBUG, "  adding subject attribute %s\n",
		XACML_ATTR_PROFILE_NS_URL_CERT_CHAIN);
	xacml_attribute_addvalue(subj_attr, subject_proxy);
	xacml_subject_addattribute(subject,subj_attr);
    }

    xacml_request_addsubject(*request,subject);

    return 0;
}

int pepc_construct_request_action     (xacml_request_t ** request,
                                       const char * actiontype,
                                       const char * actionid)
{
    const char * logstr = "pepc_construct_request_action";
    xacml_action_t * action = NULL;
    xacml_attribute_t * attr_actiontype = NULL;

    if ((!request) || (!(*request)))
    {
        lcmaps_log (LOG_ERR, "No request object presented, failure\n");
        return 1;
    }

    lcmaps_log(LOG_DEBUG, "set PEP action(actiontype)...\n");
    action = xacml_action_create();
    if (!action)
    {
        lcmaps_log (LOG_ERR, "%s: failed to create xacml_action_t object\n", logstr);
        return 1;
    }

    attr_actiontype = xacml_attribute_create(XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_ID);
    if (!attr_actiontype)
    {
        lcmaps_log (LOG_ERR, "%s: failed to create xacml_attribute_t object\n", logstr);
        return 1;
    }
    
    if (actionid)   {
	lcmaps_log(LOG_DEBUG, "  adding action attribute %s = %s\n",
		XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_ID,
		actionid);
	xacml_attribute_addvalue(attr_actiontype, actionid);
    } else {
	lcmaps_log(LOG_DEBUG, "  adding action attribute %s = %s\n",
		XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_ID,
		actiontype);
	xacml_attribute_addvalue(attr_actiontype, actiontype);
    }
    xacml_action_addattribute(action, attr_actiontype);

    xacml_request_setaction(*request,action);

    return 0;
}

/***************************************************
Function:      pepc_construct_request_resource
Description:   Sets up the resource section of the 
               request
***************************************************/
int pepc_construct_request_resource   (xacml_request_t ** request,
                                       const char * resourceid,
                                       const char * resourcetype,
                                       const char * dnshostname,
				       int use_profile)
{
    const char * logstr = "pepc_construct_request_resource";
    xacml_resource_t * resource = NULL;
    xacml_attribute_t * resource_attr   = NULL;

    if ((!request) || (!(*request)))
    {
        lcmaps_log (LOG_ERR, "No request object presented, failure\n");
        return 1;
    }

    lcmaps_log(LOG_DEBUG, "add PEP resource(resourcetype, dnshostname)...\n");
    resource = xacml_resource_create();
    if (!resource) {
        lcmaps_log (LOG_ERR, "%s: Unable to create xacml_resource_t object\n", logstr);
        return 1;
    }
    
    /* Same attribute ID in both profiles */
    resource_attr = xacml_attribute_create(XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_ID);
    if (!resource_attr) {
        lcmaps_log (LOG_ERR, "%s: failed to created xacml_attribute_t object\n", logstr);
        return 1;
    }

    /* 
     * If there is a specific resource id specified:
     * move away from the profile document and go specs compliant
     */
    if (resourceid) {
	lcmaps_log(LOG_DEBUG, "  adding resource attribute %s = %s\n",
		XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_ID,
		resourceid);
	xacml_attribute_addvalue(resource_attr, resourceid);
    } else {
	lcmaps_log(LOG_DEBUG, "  adding resource attribute %s = %s\n",
		XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_ID,
		resourcetype);
	xacml_attribute_addvalue(resource_attr, resourcetype);
    }
    xacml_resource_addattribute(resource, resource_attr);

    /* dnshostname only in interoperability profile */
    if (use_profile & PROFILE_AUTHZ_INTEROP) {
	resource_attr = xacml_attribute_create(XACML_ATTR_PROFILE_NS_URL_RESOURCE_DNS_HOST_NAME);
	if (!resource_attr) {
	    lcmaps_log (LOG_ERR, "%s: failed to create xacml_attribute_t object\n", logstr);
	    return 1;
	}
	lcmaps_log(LOG_DEBUG, "  adding resource attribute %s = %s\n",
		XACML_ATTR_PROFILE_NS_URL_RESOURCE_DNS_HOST_NAME,
		dnshostname);
	xacml_attribute_addvalue(resource_attr, dnshostname);
	xacml_resource_addattribute(resource, resource_attr);
    }

    xacml_request_addresource(*request,resource);

    return 0;
}


/*********************************************************************
Function:      pepc_construct_request_environment
Description:   Sets up the environment section of the 
               request.
               The proxy certificate used to execute the pilot job framework (not the real user job)
               will be set in the environment of the request.
*********************************************************************/
int pepc_construct_request_environment (xacml_request_t ** request,
                                        const char * environ_proxy,
					int use_profile)
{
    const char * logstr = "pepc_construct_request_environment";
    char * supp_obl_id = NULL;
    xacml_environment_t * environment = NULL;
    xacml_attribute_t * env_attr      = NULL;

    if ((!request) || (!(*request))) {
        lcmaps_log (LOG_ERR, "%s: No request object presented, failure\n", logstr);
        return 1;
    }
    
    lcmaps_log(LOG_DEBUG, "set PEP environment...\n");
    environment= xacml_environment_create();
    if (!environment) {
        lcmaps_log (LOG_ERR, "%s: failed to create xacml_environment_t object\n", logstr);
        return 1;
    }

    /* Add SWITCH Argus profile environment attribute(s) */
    if (use_profile & PROFILE_SWITCH_GRIDWN) {
	/* Add profile ID attribute */
	env_attr = xacml_attribute_create(XACML_GRIDWN_ATTRIBUTE_PROFILE_ID);
	if (!env_attr) {
            lcmaps_log (LOG_ERR, "%s: failed to create xacml_attribute_t object\n", logstr);
            return 1;
        }
	lcmaps_log(LOG_DEBUG, "  adding environment attribute %s = %s\n",
		XACML_GRIDWN_ATTRIBUTE_PROFILE_ID,
		XACML_GRIDWN_PROFILE_VERSION);
	xacml_attribute_setdatatype(env_attr,XACML_DATATYPE_ANYURI);
	xacml_attribute_addvalue(env_attr,XACML_GRIDWN_PROFILE_VERSION);
	xacml_environment_addattribute(environment,env_attr);
    } 
    
    /* Add AuthZ Interop profile environment attribute(s) */
    if (use_profile & PROFILE_AUTHZ_INTEROP) {
	/* Adding pilot proxy to environment */
	if (environ_proxy) {
	    env_attr = xacml_attribute_create(XACML_ATTR_PROFILE_NS_URL_ENVIRONMENT_PILOT_CERT_CHAIN);
	    if (!env_attr) {
		lcmaps_log (LOG_ERR, "%s: failed to create xacml_attribute_t object\n", logstr);
		return 1;
	    }
	    lcmaps_log(LOG_DEBUG, "  adding environment attribute %s\n",
		    XACML_ATTR_PROFILE_NS_URL_ENVIRONMENT_PILOT_CERT_CHAIN);
	    xacml_attribute_addvalue(env_attr, environ_proxy);
	    xacml_environment_addattribute(environment, env_attr);
	}

	/* Add supported obligations */
	rewind_supported_obligation_list();
	while ((supp_obl_id = get_next_supported_obligation_id())) {
	    env_attr = xacml_attribute_create(XACML_ATTR_PROFILE_NS_URL_ENVIRONMENT_ATTRIBUTES_OBLIG_SUPPORTED);
	    if (!env_attr) {
		lcmaps_log (LOG_ERR, "%s: failed to create xacml_attribute_t object\n", logstr);
		return 1;
	    }
	    lcmaps_log(LOG_DEBUG, "  adding environment attribute %s = %s\n",
		    XACML_ATTR_PROFILE_NS_URL_ENVIRONMENT_ATTRIBUTES_OBLIG_SUPPORTED,
		    supp_obl_id);
	    xacml_attribute_addvalue(env_attr, supp_obl_id);
	    xacml_environment_addattribute(environment, env_attr);
	}
    }

    /* Add environment object to request */
    xacml_request_setenvironment(*request,environment);

    return 0;
}



#if PEP_API == 1
int pepc_engage(const char *endpoint, unsigned int n_endpoints, int *retry)
#else
int pepc_engage(PEP *pep_handle, const char *endpoint, unsigned int n_endpoints, int *retry)
#endif
{
    xacml_decision_t decision;
    pep_error_t pep_rc = PEP_OK;
    int rc = 0;

    /* Wait for 50 milli seconds, plus a random number that
     * randomizes below the 10 milli seconds if we use the same
     * endpoint */
    if (endpoint==prev_endpoint)
	usleep (50000 + pepc_rand() % 10000);
    else    {
	/* Set correct endpoint */
	pep_rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_URL, endpoint);
	if (pep_rc != PEP_OK) {
	    lcmaps_log (LOG_ERR,
			"%s: Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_URL, %s) failed: %d: %s\n",
			__func__, endpoint, (int)pep_rc, pep_strerror(pep_rc));
	    return -1;
	}
	/* Log it (debug) when we move to a different one */
	if (prev_endpoint)
	    lcmaps_log(LOG_DEBUG, "%s: endpoint updated to %s\n",
		    __func__, endpoint);

	/* Store as previous endpoint */
	prev_endpoint=endpoint;
    }
    
    /* Make sure to reset the decision */
    pepc_reset_decision();

    /* send authz request and process */
#if PEP_API == 1
    pep_rc= pep_authorize(&pepc_request,&pepc_response);
#else
    pep_rc= pep_authorize(pep_handle, &pepc_request,&pepc_response);
#endif

    /* Log the cURL output to LCMAPS log at LOG_DEBUG level, warnings are
     * already logged */
    pep_log_pipe(LOG_DEBUG);

    /* Parse result of pep_authorize */
    if (pep_rc!=PEP_OK)	{
	lcmaps_log (LOG_ERR,
		"%s: pep_authorize() failed for endpoint %s: \"%s\"\n",
		__func__, endpoint, pep_strerror(pep_rc));

	/* retry if the error is curl related, even if the number of endpoints
	 * is only 1, since it can be a timeout */
	*retry=(pep_rc<PEP_ERR_CURL ? 0 : 1);
	rc=-1;
	goto cleanup;
    }

    /* pep_authorize succeeded: check the decision */
    if ( pepc_get_decision(&decision)!=0 ||
	 decision==XACML_DECISION_INDETERMINATE)
    {
	/* decision is invalid, retry at different endpoint */
	*retry=(n_endpoints>1 ? 1 : 0);
	rc=-1;
	goto cleanup;
    }

    /* Answer is final: do not retry */ 
    *retry=0;
    /* Check consistency of the permit result */
    if (decision==XACML_DECISION_PERMIT &&
	(unhandled_obligations(pepc_response) ||
	 parse_and_store_creddata()))
	rc=-1;

cleanup:
    /* Cleanup internal list */
    clean_uidgid_lists();

    return rc;
}




#if PEP_API == 1
int pepc_cleanup (void)
{
    pep_error_t pep_rc = PEP_OK;
#else
int pepc_cleanup (PEP *pep_handle)
{
#endif
    
	lcmaps_log(LOG_DEBUG, "delete PEP request and response...\n");
	xacml_request_delete(pepc_request);
	xacml_response_delete(pepc_response);
	/* Make sure to set all global static variables back to NULL */
	pepc_request=NULL;
	pepc_response=NULL;
	prev_endpoint=NULL;

	/* Cleanup pipe logging: flush one more time, close only the read end of
	 * the pipe, as curl will do the write end */
	pep_log_pipe(LOG_DEBUG);
	pep_log_pipe(0);

	lcmaps_log(LOG_DEBUG, "destroy PEP...\n");

#if PEP_API == 1
	pep_rc= pep_destroy();
	if (pep_rc != PEP_OK) {
		lcmaps_log (LOG_ERR, "Error: pep_destroy() failed: %d: %s\n",pep_rc,pep_strerror(pep_rc));
		return pep_rc;
	}
#else
	pep_destroy(pep_handle);
#endif

	lcmaps_log(LOG_DEBUG, "done.\n");
	return 0;
}
