/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

#ifndef _PEP_C_LOG_H
#define _PEP_C_LOG_H

#include <stdarg.h>

/* include lcmaps_c_pep_config.h for PEP_API */
#include "lcmaps_c_pep_config.h"

/* Includes for PEP-C library */
#if PEP_API == 1
#include <pep/pep.h>
#else
#include <argus/pep.h>
#endif

/**
 * Setup piped log output for cURL, otherwise it will go to stderr. Actual log
 * will be logged to lcmaps log by pep_log_curl()
 * argument pep: pep handle, used to get the curl handle
 * Returns 0 on success, -1 on error
 */
#if PEP_API == 1
int pepapi_setup_log(int pep_c_loglevel);
#else
int pepapi_setup_log(PEP *pep_handle, int pep_c_loglevel);
#endif

/**
 * Logs output from cURL, which is stored in pipe, to LCMAPS log on given level.
 * Returns 0 on success, -1 on error
 */
int pep_log_pipe(int level);

/*
 * Cleanup the pipe and associated file pointer
 * parameter closewrite: when set, close also the write end of the pipe,
 *                       otherwise only close the read end. It will still reset
 *                       the fd and stream to -1 and NULL to allow them to be
 *                       re-used. This hack is necessary since curl will close
 *                       the open filedescriptor, but we won't know it.
 * Returns 0 on success or -1 on error.
 */
void pep_close_log_pipe(int closewrite);

#endif /* _PEP_C_LOG_H */
